#ifdef __APPLE__
#include <mach/mach.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <sys/time.h>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <vector>
#include <string>
#include <map>
#include <unordered_map>
#include <cedar.h>
#include <trie.h>
#include <critbit.h> // need to modify critbit0_insert to take length and value
#include <btree_map.h>
#include <google/sparse_hash_map>
#include <google/dense_hash_map>
#pragma GCC diagnostic warning "-w"
#include <Judy.h>
#include <hat-trie/hat-trie.h>
extern "C" {
#include <splay.h>
}
#pragma GCC diagnostic warning "-Wall"

// static const
static const size_t BUFFER_SIZE = 1 << 16;
// typedef
#if   defined (USE_CEDARPP)
typedef cedar::da <int, -1, -2, false>             cedar_t;
#else
typedef cedar::da <int>                            cedar_t;
#endif
typedef dutil::trie                                trie_t;
typedef std::map <std::string, int>                map_t;
typedef std::unordered_map <std::string, int>      hash_t;
typedef critbit0_tree                              cbit_t;
typedef btree::btree_map <std::string, int>        gbtree_t;
typedef google::sparse_hash_map <std::string, int> gshash_t;
typedef google::dense_hash_map <std::string, int>  gdhash_t;
typedef Pvoid_t                                    judy_t;
typedef hattrie_t                                  hat_t;
typedef splay                                      splay_t;

size_t read_data (char* file, char*& data) {
  int fd = ::open (file, O_RDONLY);
  if (fd < 0)
    errx (1, "no such file: %s", file);
  size_t size = static_cast <size_t> (::lseek (fd, 0L, SEEK_END));
  data = new char[size];
  ::lseek (fd, 0L, SEEK_SET);
  ::read  (fd, data, size);
  ::close (fd);
  return size;
}

template <typename T>
inline T* create () { return new T (); }

template <>
inline trie_t* create () { return dutil::trie::create_trie (dutil::trie::SINGLE_TRIE); }

template <>
inline gdhash_t* create () { gdhash_t* p = new gdhash_t; p->set_empty_key (""); return p; }

template <>
inline hat_t*  create () { return hattrie_create (); }

template <>
inline splay_t* create () { return NULL; }

template <typename T>
inline void destroy (T* t) { delete t; }

template <>
inline void destroy (cbit_t* t) { critbit0_clear (t); delete t; }

template <>
inline void destroy (judy_t* t) { Word_t bytes = 0; JSLFA (bytes, *t); delete t; }

template <>
inline void destroy (hat_t* t) { hattrie_free (t); }

template <>
inline void destroy (splay_t* t) {
  if (t == NULL) return;
  destroy (t->left);
  destroy (t->right);
  std::free (t->item);
  std::free (t);
}

size_t get_process_size () {
#ifdef __APPLE__
  struct task_basic_info t_info;
  mach_msg_type_number_t t_info_count = TASK_BASIC_INFO_COUNT;
  task_info (current_task (), TASK_BASIC_INFO,
             reinterpret_cast <task_info_t> (&t_info), &t_info_count);
  return t_info.resident_size;
#else
  FILE* fp = std::fopen("/proc/self/statm", "r");
  size_t dummy (0), vm (0);
  std::fscanf (fp, "%ld %ld ", &dummy, &vm); // get resident (see procps)
  std::fclose (fp);
  return vm * ::getpagesize ();
#endif
}

template <typename T> // map/unordered_map/btree
inline void insert_key (T*& t, char* key, size_t len, int& n_)
{ t->insert (typename T::value_type (std::string (key, len), ++n_)); }
template <typename T>  // map/unordered_map/btree
inline void lookup_key (T*& t, char* key, size_t len, int& n_)
{ if (t->find (std::string (key, len)) != t->end ()) ++n_; }

// cedar
template <>
inline void insert_key <cedar_t> (cedar_t*& t, char* key, size_t len, int& n_)
{ t->update (key, len) = ++n_; }
template <>
inline void lookup_key <cedar_t> (cedar_t*& t, char* key, size_t len, int& n_)
{ if (t->exactMatchSearch <int> (key, len) >= 0) ++n_; }

// libtrie
template <>
inline void insert_key <trie_t> (trie_t*& t, char* key, size_t len, int& n_)
{ t->insert (key, len, ++n_); }
template <>
inline void lookup_key <trie_t> (trie_t*& t, char* key, size_t len, int& n_)
{ int val = 0; if (t->search (key, len, &val)) ++n_; }

// crit-bit tree
// modified critbit0_insert so that it stores an integer value after the string
template <>
inline void insert_key <cbit_t> (cbit_t*& t, char* key, size_t len, int& n_)
{ critbit0_insert (t, key, len, ++n_); }
template <>
inline void lookup_key <cbit_t> (cbit_t*& t, char* key, size_t len, int& n_)
{ if (critbit0_contains (t, key, len)) ++n_; }

// judy array
template <>
inline void insert_key <judy_t> (judy_t*& t, char* key, size_t len, int& n_)
{ Word_t* PValue = 0; JSLI (PValue, *t, key); *PValue = ++n_; }
template <>
inline void lookup_key <judy_t> (judy_t*& t, char* key, size_t len, int& n_)
{ PWord_t PValue = 0; JSLG (PValue, *t, key); if (PValue) ++n_; }

// hat trie
template <>
inline void insert_key <hat_t> (hat_t*& t, char* key, size_t len, int& n_)
{ *hattrie_get (t, key, len) = ++n_; }
template <>
inline void lookup_key <hat_t> (hat_t*& t, char* key, size_t len, int& n_)
{ if (hattrie_tryget (t, key, len)) ++n_; }

// splay tree
// store an integer value after the string as in crit-bit tree
template <>
inline void insert_key <splay_t> (splay_t*& t, char* key, size_t len, int& n_) {
  char* data = reinterpret_cast <char*> (std::malloc (len + 1 + sizeof (int)));
  std::strcpy (data, key);
  std::memcpy (data + len + 1, &(++n_), sizeof (int));
  t = splay_insert (t, data, (cmpfunc)std::strcmp);
}
template <>
inline void lookup_key <splay_t> (splay_t*& t, char* key, size_t len, int& n_)
{ if (reinterpret_cast <char*> (splay_find (&t, key, (cmpfunc)std::strcmp))) ++n_; }

#ifdef USE_BINARY_DATA
inline bool is_splitter (const char p) { return p == '\0'; }
#else
inline bool is_splitter (char &p) { if (p != '\n') return false; p = '\0'; return true; }
#endif

template <typename T>
void insert_data (T*& t, int fd, int& n, int& n_) {
  char data[BUFFER_SIZE];
  ssize_t start (0), end (0);
  while (ssize_t size = ::read (fd, data + end, BUFFER_SIZE - end) + end) {
    for (; end < size; ++end)
      if (is_splitter (data[end]))
        ++n, insert_key (t, data + start, end - start, n_), start = end + 1;
    std::memmove (data, data + start, size - start);
    end = size - start; start = 0;
  }
}

template <typename T>
void lookup_data (T*& t, char* data, size_t size, int& n, int& n_) {
  for (size_t start (0), end (0); end < size; ++end)
    if (is_splitter (data[end]))
      ++n, lookup_key (t, data + start, end - start, n_), start = end + 1;
}

template <typename T>
void bench (const char* label, char* keys, char* queries) {
  T* t = 0;
  int n (0), n_(0);
  double elapsed = 0;
  char* data = 0;
  struct timeval start, end;
  {
    // load keys
    size_t rss = get_process_size ();
    // build trie
    t = create <T> ();
    gettimeofday (&start, NULL);
    int fd = ::open (keys, O_RDONLY);
    if (fd < 0)
      errx (1, "no such file: %s", keys);
    insert_data (t, fd, n, n_);
    close (fd);
    gettimeofday (&end, NULL);
    elapsed = end.tv_sec + end.tv_usec * 1e-6 - (start.tv_sec + start.tv_usec * 1e-6);
    std::fprintf (stderr, "---- %-7s ---------------------------------------\n", label);
    std::fprintf (stderr, "Init RSS:            %.2f MiB (%ld bytes)\n", rss / 1048576.0, rss);
    std::fprintf (stderr, "Time to insert:      %.2f sec (%.2f nsec per key)\n",
                  elapsed, elapsed * 1e9 / n);
    std::fprintf (stderr, "Words:               %d\n\n", n);
  }
  // load data
  if (std::strcmp (queries, "-") != 0) {
    size_t size = read_data (queries, data);
    // search
    n = n_ = 0;
    gettimeofday (&start, NULL);
    lookup_data (t, data, size, n, n_);
    gettimeofday (&end, NULL);
    elapsed = end.tv_sec + end.tv_usec * 1e-6 - (start.tv_sec + start.tv_usec * 1e-6);
    std::fprintf (stderr, "Time to search:      %.2f sec (%.2f nsec per key)\n",
                  elapsed, elapsed * 1e9 / n);
    std::fprintf (stderr, "Words:               %d\n", n);
    std::fprintf (stderr, "Found:               %d\n", n_);
    delete [] data;
  }
  if (t) destroy (t);
}

int main (int argc, char** argv) {
  if (argc < 3)
    ::errx (1, "Usage: %s keys test", argv[0]);
  //
#ifdef USE_CEDAR
  bench <cedar_t>  ("cedar",   argv[1], argv[2]);
#endif
#ifdef USE_CEDARPP
  bench <cedar_t>  ("cedar++", argv[1], argv[2]);
#endif
#ifdef USE_LIBTRIE
  bench <trie_t>   ("libtrie", argv[1], argv[2]);
#endif
#ifdef USE_MAP
  bench <map_t>    ("map",     argv[1], argv[2]);
#endif
#ifdef USE_HASH
  bench <hash_t>   ("hash",    argv[1], argv[2]);
#endif
#ifdef USE_CBIT
  bench <cbit_t>   ("cbit",    argv[1], argv[2]);
#endif
#ifdef USE_GOOGLE_BTREE
  bench <gbtree_t> ("gbtree",  argv[1], argv[2]);
#endif
#ifdef USE_GOOGLE_SHASH
  bench <gshash_t> ("gshash",  argv[1], argv[2]);
#endif
#ifdef USE_GOOGLE_DHASH
  bench <gdhash_t> ("gdhash",  argv[1], argv[2]);
#endif
#ifdef USE_JUDY
  bench <judy_t>   ("judy",    argv[1], argv[2]);
#endif
#ifdef USE_HAT
  bench <hat_t>    ("hat",     argv[1], argv[2]);
#endif
#ifdef USE_SPLAY
  bench <splay_t>  ("splay",   argv[1], argv[2]);
#endif
}
/*
  gcc -O2 -g -c splay.c
  gcc -O2 -g -std=c99 -c critbit.c
  
  g++ -fpermissive -std=c++11 -I. -I$HOME/local/include -O2 -g critbit.o splay.o bench.cc -o bench -L$HOME/local/lib -lhat-trie -lJudy -ltrie
  g++ -DUSE_BINARY_DATA -fpermissive -std=c++11 -I. -I$HOME/local/include -O2 -g critbit.o splay.o bench.cc -o bench_bin -L$HOME/local/lib -lhat-trie -lJudy -ltrie 
*/
